# An experiment to verify the logic in https://www.youtube.com/watch?v=iSNsgj1OCLA
# 100 prisoners, each with a number, enters a room with 100 numbered boxes, each with a
# random number inside. Each may open only 50 boxes. If everyone finds their own number,
# they are all freed. If anyone fails to find their number, they are all executed. There
# is no communication once the game starts. Can they do better than 0.5**100 chance?
# This model verifies that indeed they can.

#===============
# SETUP
#===============

from random import shuffle, randint
from helipad import Helipad
from numpy import mean

heli = Helipad()
heli.name = 'Prison Escape'
heli.agents.order = 'random'

heli.params.add('strategy', 'Optimal Strategy', 'check', dflt=True)
heli.param('num_agent', 100)
heli.params['num_agent'].type = 'hidden'
heli.param('stopafter', 500)
heli.cumAvg = 0

#===============
# BEHAVIOR
#===============

#Set up boxes. Index is the box number, value is the number inside.
@heli.hook
def modelPreStep(model):
	model.boxes = [a.id for a in model.agents['agent']]
	shuffle(model.boxes)
	model.escaped = True

@heli.hook
def agentStep(agent, model, stage):
	agent.checked = []

	#Choose the box with your own number, then choose the box with the number
	#you find inside, and so on until you find your own number or run out.
	if model.param('strategy'):
		agent.checked.append(model.boxes[agent.id-1])
		while agent.id not in agent.checked and len(agent.checked) < 50:
			agent.checked.append(model.boxes[agent.checked[-1]-1])

	#Choose randomly for 50 boxes
	else:
		while agent.id not in agent.checked and len(agent.checked) < 50:
			box = randint(0, len(model.boxes)-1)
			while box in agent.checked: box = randint(0, len(model.boxes)-1)
			agent.checked.append(model.boxes[box])

	if agent.id not in agent.checked:
		model.escaped = False
		model.cutStep() #Go ahead and skip everyone else if anyone fails

@heli.hook
def modelPostStep(model):
	model.cumAvg = mean(model.data['escaped']+[model.escaped])

#===============
# DATA AND VISUALIZATION
#===============

from helipad.visualize import TimeSeries
viz = heli.useVisual(TimeSeries)
viz.dim = (950, 400)

heli.data.addReporter('escaped', heli.data.modelReporter('escaped'))
heli.data.addReporter('cumAvg', heli.data.modelReporter('cumAvg'))

caplot = viz.addPlot('cumAvg', 'Cumulative Average')
caplot.addSeries('cumAvg', 'Cumulative Average', '#C00')

#===============
# LAUNCH THE GUI
#===============

heli.launchCpanel()


"""
The control panel class for use in Jupyter notebooks. This module should not be imported directly; use `model.launchCpanel()` instead.
"""

import os
from collections import ChainMap
from ipywidgets import interactive, Layout, Accordion, HBox, VBox, HTML, Label, Button, FloatProgress
from IPython.display import display
from helipad.param import Param
from helipad.helpers import ï

class Cpanel(VBox):
	"""The control panel class for use in Jupyter notebooks. https://helipad.dev/functions/cpanel/"""
	def __init__(self, model, redraw: bool=False):
		super().__init__()
		self.model = model
		if redraw:
			self.children = ()
			self.remove_class('invalid')
		else: self.add_class('helipad_cpanel')
		self.valid = True

		#CSS niceties
		__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
		with open(os.path.join(__location__,'ipy-styles.css'), encoding='UTF-8') as c: css = c.read()

		#CSS for goods and breeds, since Ipywidgets ≥8.0 oversanitizes HTML in description attributes
		for n,c in ChainMap(*[{f'breed_{p}_{k}': v.color.hex for k,v in d.breeds.items()} for p,d in model.agents.items()]+[{'good_'+k: v.color.hex for k,v in model.goods.items()}]).items():
			css += f'.helipad_{n} .widget-label::before {{ background: {c} }}'

		self.children += HTML(value='<style type="text/css">'+css+'</style>'),

		#Callback function generator for Jupyter elements
		#Attaches to the `param` object, so the param argument is a `self`
		def setVar(param, item=None):
			def sv(val): #Ipywidgets bases on the function signature, so can't use more than one here…
				if param.type=='checkgrid': param.set(item, val, updateGUI=False)
				else: param.set(val, item, updateGUI=False)

				if callable(param.callback):
					if param.per is None:
						pval = param.get(item) if param.type != 'checkgrid' else (item, param.get(item))
						param.callback(self.model, param.name, pval)
					else: param.callback(self.model, param.name, item, param.get(item))

			#Consolidate value from bool and string, and toggle entry disabled state
			#Have to return a different function since Ipywidgets bases the interactive off the function signature
			if param.type=='checkentry':
				def sv2(b: bool, s: str):
					#Ipywidgets ≥8.0 runs the callback before the element is assigned
					try:
						els = param.element if item is None else param.elements[item]
						els.children[1].disabled = not b
					except (AttributeError, KeyError): return

					#Coercing an int can fail, so if there's an exception, reset the textbox content
					try:
						val = s if (b and s=='') or 'func〈' in s else (param.entryType(s) if b else False)
						sv(val)
					except: els.children[1].value = str(param.get())

				return sv2
			else: return sv
		Param.setVar = setVar

		ctop = self.model.doHooks('CpanelTop', [self, None])
		if ctop: self.children += ctop,

		#Global config
		for n,param in model.params.items():
			if not getattr(param, 'config', False) or param.type=='checkgrid': continue
			param.element = self.renderParam(param)
			if param.element is not None: self.children += (param.element,)
			if param.name=='csv': param.set(ï('filename'))
			if n=='stopafter' and not param.event and not param.get(): param.element.children[1].value = '10000'
			if param.type=='checkentry' and getattr(param, 'config', False) and not (n=='stopafter' and param.event): param.set(False)

		caip = self.model.doHooks('CpanelAboveItemParams', [self, None])
		if caip: self.children += caip,

		#Per-good parameters
		for param in model.params.perGood.values():
			param.element = self.renderParam(param)
			self.children += param.element,

		#Per-breed parameters
		for param in model.params.perBreed.values():
			param.element = self.renderParam(param)
			self.children += param.element,

		cap = self.model.doHooks('CpanelAboveParams', [self, None])
		if cap: self.children += cap,

		#Pull out grouped parameters
		groups = []
		for group in self.model.params.groups: groups += list(group.members.keys())

		#Global parameters
		for k, param in model.params.globals.items():
			if getattr(param, 'config', False) or k in groups or param.type=='checkgrid': continue
			param.element = self.renderParam(param)
			if param.element is not None: self.children += param.element,

		#Param groups
		for group in model.params.groups:
			for param in group.members.values():
				param.element = self.renderParam(param)
			group.element = Accordion(children=[HBox([p.element for p in group.members.values()])], selected_index=0 if group.open else None)
			group.element.set_title(0, group.title)
			group.element.add_class('helipad_paramgroup')
			self.children += group.element,

		#Checkgrids
		for param in model.params.values():
			if param.type!='checkgrid' or param.name=='shocks': continue
			param.element = self.renderParam(param)
			if param.element is not None: self.children += param.element,

		cas = self.model.doHooks('CpanelAboveShocks', [self, None])
		if cas: self.children += cas,

		#Shocks
		if len(model.shocks):
			def sfuncButton(slef, *args): slef.do(self.model)
			model.shocks.Shock.setButton = sfuncButton
			model.params['shocks'].elements = {}
			children = []
			for shock in model.shocks.shocksExceptButtons.values():
				shock.element = interactive(shock.setCallback, val=shock.selected)
				shock.element.children[0].description = shock.name
				shock.element.children[0].description_tooltip = shock.desc if shock.desc is not None else ''
				children.append(shock.element)
				model.params['shocks'].elements[shock.name] = shock.element #For setting via model.param()
			buttons = []
			for shock in model.shocks.buttons.values():
				shock.element = Button(description=shock.name, icon='bolt')
				shock.element.click = shock.setButton
				shock.element.description_tooltip = shock.desc if shock.desc is not None else ''
				buttons.append(shock.element)
			children.append(HBox(buttons))
			model.params['shocks'].element = Accordion(children=[VBox(children)], selected_index=0)
			model.shocks.element = model.params['shocks'].element
			model.params['shocks'].element.set_title(0, model.params['shocks'].title)
			self.children += model.params['shocks'].element,

		cbot = self.model.doHooks('CpanelBottom', [self, None])
		if cbot: self.children += cbot,

		self.postinstruct = self.displayAlert(ï('After setting parameter values, run <code>launchVisual()</code> or <code>start()</code> to start the model.'))
		if not redraw:
			display(self)

			class progressBar(FloatProgress):
				def __init__(self):
					super().__init__(min=0, max=1)

				def determinate(self, det: bool):
					self.mode = 'determinate' if det else 'indeterminate'
					if det: self.remove_class('indeterminate')
					else:
						self.add_class('indeterminate')
						self.value = 1

				def update(self, n): self.value = n
				def start(self): self.add_class('helipad_running')
				def stop(self): self.remove_class('helipad_running')
				def done(self): self.layout.visibility = 'hidden'

			class runButton(Button):
				def __init__(self2, **kwargs):
					super().__init__(**kwargs)
					self2.click = self.model.stop

				def run(self2):
					self2.click = self.model.stop
					self2.description = ï('Pause')
					self2.icon = 'pause'

				def pause(self2):
					self2.click = self.model.start
					self2.description = ï('Run')
					self2.icon = 'play'

				def terminate(self):
					self.layout.visibility = 'hidden'

			#Remove previous hooks so we don't double up when re-running launchCpanel()
			model.hooks.remove('modelPreSetup', 'cpanel_visualPreLaunch')
			model.hooks.remove('terminate', 'cpanel_terminate')

			#Model flow control: pause/run button and progress bar
			@model.hook('modelPreSetup', prioritize=True)
			def cpanel_visualPreLaunch(model):
				self.runButton = runButton(description=ï('Pause'), icon='pause')
				self.progress = progressBar()
				self.postinstruct.layout = Layout(display='none')

				self.stopbutton = Button(description=ï('Stop'), icon='stop')
				self.stopbutton.click = self.model.terminate

				pbararea = HBox([self.runButton, self.stopbutton, self.progress])
				pbararea.add_class('helipad_progress_area')
				display(pbararea)

			@model.hook('terminate')
			def cpanel_terminate(model, data):
				self.postinstruct.layout = Layout(display='inline-block')
				self.stopbutton.layout.visibility = 'hidden'

	def renderParam(self, param, item=None):
		"""Constructs an Ipywidget from a `Param` object. https://helipad.dev/functions/cpanel/renderparam/"""
		i=None

		#Generate an accordion for per-item parameters
		if param.per is not None and item is None:
			param.elements = {}
			for name, good in param.pKeys.items():
				if getattr(good, 'money', False): continue
				param.elements[name] = self.renderParam(param, item=name)

			accordion = Accordion(children=[HBox(list(param.elements.values()))], selected_index=0)
			accordion.set_title(0, param.title)
			accordion.add_class('helipad_param_peritem helipad_paramgroup')
			return accordion

		elif param.type=='slider':
			if isinstance(param.opts, dict): i = interactive(param.setVar(item), val=(param.opts['low'],param.opts['high'], param.opts['step']))
			else:
				s = interactive(param.setVar(item), val=(0, len(param.opts)-1, 1))
				s.children[0].readout = False
				l = Label(value=str(param.opts[0]), layout=Layout(margin='0 0 0 15px'))
				i = HBox([s.children[0],l])

		elif param.type=='check':
			i = interactive(param.setVar(item), val=param.get(item))
		elif param.type=='menu':
			i = interactive(param.setVar(item), val=[(k[1], k[0]) for k in param.opts.items()])
		elif param.type=='checkentry':
			val = param.get(item)
			defaults = (
				(isinstance(val, param.entryType) or val) if not (param.name=='stopafter' and param.event) else True,				#Bool
				(str(val) if isinstance(val, param.entryType) else '') if not (param.name=='stopafter' and param.event) else 'Event: '+val	#Str
			)
			i = interactive(param.setVar(item), b=defaults[0], s=defaults[1])
			if param.per is None:
				i = HBox(i.children)
				i.children[0].layout = Layout(width='150px')
			if val is False: i.children[1].disabled = True
			i.children[1].description = ''

			if param.name=='stopafter' and param.event:
				i.children[0].disabled = True
				i.children[1].disabled = True
				i.add_class('helipad_checkentry_func')
		elif param.type=='checkgrid':
			param.elements = {}
			for k,v in param.opts.items():
				if not isinstance(v, (tuple, list)): v = (v, None)
				elif len(v) < 2: v = (v[0], None)
				param.elements[k] = interactive(param.setVar(k), val=param.vars[k])
				param.elements[k].children[0].description = v[0]
				param.elements[k].children[0].description_tooltip = v[1] if v[1] is not None else '' #Not working, not sure why

			#Toggle-all button
			#Would like to put it in .p-Collapse-header, but I don't think I can place an element in there
			#So append it to the children and we'll absolutely position it to the right place
			param.elements['toggleAll'] = Button(icon='check')
			def toggleAll(b=None):
				v = not [c.children[0].value for c in param.element.children[0].children if not isinstance(c, Button) and c.children[0].value]
				for c in param.element.children[0].children:
					if not isinstance(c, Button) and not c.children[0].disabled: c.children[0].value = v
			param.elements['toggleAll'].on_click(toggleAll)
			param.elements['toggleAll'].add_class('helipad_toggleAll')

			param.containerElement = HBox(list(param.elements.values()))
			i = Accordion(children=[param.containerElement], selected_index=0)
			i.set_title(0, param.title if not item else item.title())
			i.add_class('helipad_checkgrid')

		if i is not None and param.type!='checkgrid':
			title = param.title if not item else item.title()
			if param.per is not None:
				n = (f'{param.prim}_' if param.per=='breed' else '')+title.lower()
				i.children[0].add_class(f'helipad_{param.per}_{n}')
				i.children[0].add_class('helipad_per_item')
			i.children[0].description = title
			i.children[0].style = {'description_width': 'initial'} #Don't truncate the label
			i.children[0].description_tooltip = param.desc if param.desc is not None else ''
			val = param.get(item)
			if param.type=='slider' and isinstance(param.opts, list):
				i.children[0].add_class('widget-logslider')
				i.children[1].value = str(val)
				if val in param.opts: val=param.opts.index(val)
			if param.type!='checkentry': i.children[0].value = val

		return i

	def displayAlert(self, text: str, inCpanel: bool=True):
		"""Display an alert element in the Jupyter notebook."""
		element = HTML(value=text)
		element.add_class('helipad_info') #Latter applies some built-in styles to the contents
		if inCpanel: self.children += element,
		else: display(element)
		return element

	def invalidate(self, message: str=ï('Model parameters changed, please re-launch the control panel with launchCpanel().')):
		"""Prevent the user from interacting with a control panel. A control panel is invalidated when another is launched."""
		self.valid = False
		self.add_class('invalid')
		warning = Label(value=message)
		warning.add_class('helipad_modal')
		self.children += warning,
		for p in self.model.params.values(): p.element = None
		return warning

#https://stackoverflow.com/questions/24005221/ipython-notebook-early-exit-from-cell
class SilentExit(Exception):
	def _render_traceback_(self): pass


"""
Helper classes and functions used internally in Helipad. This module should not be imported directly.
"""

import warnings
from sys import __stdout__
from io import BufferedWriter

#Using _ is a disaster. Can't install to global scope because it conflicts with readline;
#Can't name it _ here because `import *` skips it
def ï(text) -> str:
	"""Internationalization. Named so as to avoid a conflict with `_` in the REPL console."""
	return helipad_gettext(text)

def isIpy() -> bool:
	"""Check for any Ipython environment, including Spyder, for event loop purposes."""
	try:
		__IPYTHON__
		return True
	except NameError: return False

def isNotebook() -> bool:
	"""Check whether Helipad is running in an interactive notebook."""
	#get_ipython() comes back undefined inside callbacks. So cache the value once, the first time it runs.
	#Can try @functools.cache when Python 3.9 is required
	if '__helipad_ipy' not in globals():
		try:
			globals()['__helipad_ipy'] = 'InteractiveShell' in get_ipython().__class__.__name__
		except NameError: globals()['__helipad_ipy'] = False

	return __helipad_ipy

def isBuffered() -> bool:
	"""Check whether the current Python script is running in a buffered or unbuffered console."""
	return isinstance(__stdout__.buffer, BufferedWriter)

class Item:
	"""A generic extensible item class to store structured data. Kwargs are stored as object properties."""
	def __init__(self, **kwargs):
		for k,v in kwargs.items():
			setattr(self, k, v)

class funcStore(dict):
	"""A generic dict-like container class. https://helipad.dev/functions/funcstore/"""
	multi = False

	def add(self, name, function):
		if self.multi:
			if name not in self: self[name] = [function]
			else: self[name].append(function)
		else: self[name] = function
		return function

	def remove(self, name, fname=None, removeall=False):
		if isinstance(name, (list, tuple)): return [self.remove(n) for n in name]
		if name not in self: return False
		if self.multi and fname:
			if isinstance(fname, (list, tuple)): return [self.remove(name, f, removeall) for f in fname]
			did = False
			for h in self[name]:
				if h.__name__ == fname:
					self[name].remove(h)
					if not removeall: return True
					else: did = True
			return did
		else:
			del self[name]
			return True

import colorsys, matplotlib.colors as mplcolor
class Color:
	"""Defines a color and provides functions to manipulate it. https://helipad.dev/functions/color/"""
	def __init__(self, color):
		#Can take a hex string, color name, or [r,g,b] list/tuple.
		self.rgb = mplcolor.hex2color(color) if isinstance(color, str) else list(color)

	def __repr__(self): return f'<Color: ({round(self.r*100)},{round(self.g*100)},{round(self.b*100)})>'

	@property
	def hex(self): return mplcolor.to_hex(self.rgb)
	@property
	def hsv(self): return list(mplcolor.rgb_to_hsv(self.rgb))
	@property
	def r(self): return self.rgb[0]
	@property
	def g(self): return self.rgb[1]
	@property
	def b(self): return self.rgb[2]
	@property
	def h(self): return self.hsv[0]
	@property
	def s(self): return self.hsv[1]
	@property
	def v(self): return self.hsv[2]

	def lighten(self, factor=3):
		"""Creates a new color, lighter than the existing color by `factor`."""
		hls = colorsys.rgb_to_hls(*self.rgb)
		return Color(colorsys.hls_to_rgb(hls[0], (1-1/factor) + hls[1]/factor, hls[2]))

	def darken(self):
		"""Create a new color, slightly darker than the original."""
		hls = colorsys.rgb_to_hls(*self.rgb)
		return Color(colorsys.hls_to_rgb(hls[0], hls[1]-0.075 if hls[1]>0.075 else 0, hls[2]))

	def blend(self, color2):
		"""Create a new color halfway between the existing color and another. https://helipad.dev/functions/color/blend/"""
		return Color(((self.r+color2.r)/2, (self.g+color2.g)/2, (self.b+color2.b)/2))

def makeDivisible(n, div, c='min'):
	return n-n%div if c=='min' else n+(div-n%div if n%div!=0 else 0)


# Replicates the repeated PD tournament from Axelrod (1980)
# Paper here: https://www.jstor.org/stable/173932
# Strategies indexed here: https://axelrod.readthedocs.io/en/stable/reference/overview_of_strategies.html#axelrod-s-first-tournament
# Strategies requiring chi-squared tests for randomness or such not yet implemented
# Difference from the original: The tournament matches stochastically, rather than being a round-robin, and repeats.

#===============
# SETUP
#===============

from helipad import Helipad
from numpy import mean, random

heli = Helipad()
heli.name = 'Axelrod Tournament'
heli.agents.order = 'match'

heli.params['num_agent'].type = 'hidden' #So we can postpone breed determination until the end

#Initial parameter values match the payoffs in Table 1
heli.params.add('cc', 'C-C payoff', 'slider', dflt=3, opts={'low': 0, 'high': 10, 'step': 0.5})
heli.params.add('dc', 'D-C payoff', 'slider', dflt=5, opts={'low': 0, 'high': 10, 'step': 0.5})
heli.params.add('cd', 'C-D payoff', 'slider', dflt=0, opts={'low': 0, 'high': 10, 'step': 0.5})
heli.params.add('dd', 'D-D payoff', 'slider', dflt=1, opts={'low': 0, 'high': 10, 'step': 0.5})
heli.params.add('rounds', 'Rounds per period', 'slider', dflt=200, opts={'low': 10, 'high': 1000, 'step': 10})
heli.params.add('n', 'Agents per strategy', 'slider', dflt=3, opts={'low': 1, 'high': 10, 'step': 1}, runtime=False)

@heli.button('Reset Wealth')
def reset(model):
	for a in model.agents['agent']: a.stocks['payoff'] = 0

heli.goods.add('payoff','#009900')

#Store associated colors and default-checked state to cycle over later
strategies = {
	'alwaysCooperate': (True, '#00CC00'),
	'alwaysDefect': (True, '#CC0000'),
	'randomly': (True, '#006600'),
	'TFT': (True, '#CC9900', 'Cooperates on the first round and thereafter plays the move last played by the opponent.'),
	'Tullock': (False, '#663311', 'Cooperates for the first 11 rounds then randomly cooperates 10% less often than the opponent has in the previous 10 rounds.'),
	'Nydegger': (False, '#003399', 'A stooge which has a memory and appears to be trustworthy, potentially cooperative, but not gullible.'),
	'Grofman': (False, '#9999CC', 'Cooperates with probability 2/7 if the players did different things on the previous move, otherwise cooperates.'),
	'Shubik': (False, '#666600', 'Punishes defection with an incrementing string of retaliations.'),
	'Grudger': (True, '#000066', 'Cooperates until the opponent defects, and then defects thereafter'),
	'Davis': (False, '#660000', 'Cooperates for 10 periods, and then plays Grudger.'),
	'Feld': (False, '#000000', 'Plays tit-for-tat, but with P[Cooperate] declining to 0.5 as the rounds approach the end'),
	'Joss': (False, '#CCCCCC', 'Tit-For-Tat, except cooperation is only with 90% probability'),
}

#Build the strategies toggle
heli.params.add('strategies', 'Strategies', 'checkgrid',
	opts = {k: (k, v[2] if len(v)>2 else None) for k,v in strategies.items()},
	dflt = [k for k,v in strategies.items() if v[0]],
	runtime=False, columns=2
)

#===============
# STRATEGIES
# Return True to cooperate, and False to defect.
#===============

def TFT(rnd, history, own):
	if not history: return True
	else: return history[-1]

def alwaysCooperate(rnd, history, own): return True

def alwaysDefect(rnd, history, own): return False

def randomly(rnd, history, own): return random.choice(2, 1)[0]

def Tullock(rnd, history, own):
	if len(history) < 11: return True
	rate = mean(history[-10:])-.1
	rate = max(rate, 0)
	return random.choice(2, 1, p=[1-rate, rate])[0]

def Nydegger(rnd, history, own):
	if len(history) <= 2: return TFT(rnd, history, own)
	if len(history) == 3:
		if not history[0] and not own[1] and history[1]: return False
		else: return TFT(rnd, history, own)

	def a(i):
		n=0
		if not history[-i]: n += 2
		if not own[-i]: n += 1
		return n
	A = 16*a(1) + 4*a(2) + a(3)
	return A not in [1,6,7,17,22,23,26,29,30,31,33,38,39,45,49,54,55,58,61]

def Grofman(rnd, history, own):
	if not history: return True
	elif history[-1] ^ own[-1]: return random.choice(2, 1, p=[5/7, 2/7])[0]
	else: return True

def Shubik(rnd, history, own):
	k=0 #The number of rounds to retaliate
	rs=0 #the number of rounds I've already retaliated so far
	for r, me in enumerate(own):
		if me and not history[r]: k += 1 #Increment k every time the opponent defected while I wasn't retaliating
		if not me: rs += 1
		else: rs = 0 #Make sure we're only counting current retaliations

	if not history: return True
	if not rs and not history[-1]: return False #Start retaliation if the opponent defected last period
	if rs and rs < k: return False #Keep retaliating if you've started a string of defections and haven't reached the limit
	return True

def Grudger(rnd, history, own):
	return False not in history

def Davis(rnd, history, own):
	if len(history) < 10: return True
	else: return Grudger(rnd, history, own)

def Feld(rnd, history, own):
	if not history: return True
	if not history[-1]: return False

	pDef = rnd/heli.param('rounds')/2
	return random.choice(2, 1, p=[pDef, 1-pDef])[0]

def Joss(rnd, history, own):
	if not history or history[-1]: return random.choice(2, 1, p=[0.1, 0.9])[0]
	if not history[-1]: return False

#===============
# MODEL LOGIC
#===============

@heli.hook
def match(agents, primitive, model, stage):
	h1, h2 = ([], [])
	for rnd in range(int(model.param('rounds'))):
		#Play
		response1 = globals()[agents[0].breed](rnd, h2, h1)
		response2 = globals()[agents[1].breed](rnd, h1, h2)
		h1.append(response1)
		h2.append(response2)

		#Payoffs
		agents[0].stocks['payoff'] += model.param(('c' if response1 else 'd') + ('c' if response2 else 'd'))
		agents[1].stocks['payoff'] += model.param(('c' if response2 else 'd') + ('c' if response1 else 'd'))

#Visualization
from helipad.visualize import TimeSeries
viz = heli.useVisual(TimeSeries)
plot = viz.addPlot('payoffs', 'Payoffs')

#Add breeds last-minute so we can toggle them in the control panel
@heli.hook
def modelPreSetup(model):

	#Clear breeds from the previous run
	for b in model.agents['agent'].breeds:
		model.data.removeReporter(b+'-proportion')
	model.agents['agent'].breeds.clear()

	for k in model.param('strategies'):
		model.agents.addBreed(k, strategies[k][1])

	model.param('num_agent', len(model.agents['agent'].breeds)*model.param('n')) #Three of each strategy, for speed

	for b, d in model.agents['agent'].breeds.items():
		model.data.addReporter(b+'-proportion', proportionReporter(b))
		plot.addSeries(b+'-proportion', b, d.color)

#===============
# CONFIGURATION
#===============

#Tally up the total payoff once each period, so we don't have to do it on every agent
@heli.hook
def dataCollect(data, t):
	data.model.totalP = data.agentReporter('stocks', good='payoff', stat='sum')(data.model)

def proportionReporter(breed):
	def func(model):
		return model.data.agentReporter('stocks', breed=breed, good='payoff', stat='sum')(model)/model.totalP
	return func

#===============
# LAUNCH THE GUI
#===============

heli.launchCpanel()


# A template for bootstrapping new models.
# See https://helipad.dev/ for complete documentation

#===============
# SETUP
# Instantiate the object and add parameters, breeds, and goods below.
#===============

from helipad import Helipad
# from utility import CobbDouglas

heli = Helipad()
heli.name = 'Model Name'
heli.agents.order = 'random' #Can be changed to 'linear' or 'match'
heli.stages = 1 #Change to create a multi-stage model

# heli.params.add('name', 'title', 'type (slider, menu, or check)', dflt=default, opts={depends on type})
# heli.goods.add('good1','hex color', lambda breed: endowment)
# heli.agents.addBreed('name1', 'hex color')
# heli.agents.addBreed('name2', 'hex color')

#===============
# BEHAVIOR
# A list of hooks and their function signatures can be found at http://helipad-docs.nfshost.com/hooks/
#===============

#Any variables or properties the agent should keep track of should have default values set here.
@heli.hook
def agentInit(agent, model):
	#agent.myAgentProperty = 0
	#agent.utility = CobbDouglas(['good1'])
	pass

#Any global variables that should be kept track of should have default values set here.
@heli.hook
def modelPostSetup(model):
	#model.myModelProperty = 0
	pass

#Agent logic should be written here.
@heli.hook
def agentStep(agent, model, stage):
	pass	

#Any global code to be run each period should be hooked to modelStep, modelPreStep, or modelPostStep.
#modelStep will run as many times per period as there are stages. modelPreStep and modelPostStep
#will run at the beginning and end of each period, respectively, and do not take a stage argument.
@heli.hook
def modelStep(model, stage):
	pass

#===============
# DATA AND VISUALIZATION
# Register reporters, plots, and series here
#===============

from helipad.visualize import TimeSeries
viz = heli.useVisual(TimeSeries)

#Reporters collect data from the model each period, generally from parameters set in agentInit and modelPostSetup.

# heli.data.addReporter('myReporter1', heli.data.agentReporter('myAgentProperty', 'agent', stat='mean'))
# heli.data.addReporter('myReporter2', heli.data.modelReporter('myModelProperty'))

#Plots are areas on the graph where series can be drawn to keep track of reporter data in real time.

myplot = viz.addPlot('myplot', 'Custom Properties', logscale=False, selected=True)

#Series draw reporter data on a plot. Here we draw two series on the same plot.
#The Plot object can also be accessed later with heli['myplot'], so the following two methods are identical.

# myplot.addSeries('myplot', 'myReporter1', 'My Agent Property', 'hex color')
# heli['myplot'].addSeries('myplot', 'myReporter2', 'My Model Property', 'hex color')

#===============
# LAUNCH THE GUI
#===============

heli.launchCpanel()